--
-- AlternativeTippingShovel
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history		v1.0 - 2013-07-09 - Initial implementation
-- @history		v1.1 - 2014-06-08 - Tipping feature added
--
--

AlternativeTippingShovel = {};

function AlternativeTippingShovel:shovelLoad(xmlFile)
	local node = getXMLString(xmlFile, "vehicle.alternativeTipping#loadPoint");
	self.altTipRayDistance = getXMLFloat(xmlFile, "vehicle.alternativeTipping#rayDistance");
	
	if node ~= nil then
		node = Utils.indexToObject(self.components, node);
	else
		node = self.shovelTipReferenceNode;
	end;
	
	self.altTipPoint = node;
	self.forceShovelFill = false;
	
	if self.altTipPoint == nil then
		if self.shovelNodes ~= nil then
			for n in pairs(self.shovelNodes) do
				self.altTipPoint = n;
				break;
			end;
		else
			if not SpecializationUtil.hasSpecialization(ArticulatedAxis, self.specializations) then
				print("Alternative Tipping - Warning: It's strongly suggested that you add an shovelTipReferenceNode or shovelNode to " .. tostring(self.customEnvironment));
			end;
		end;
	end;
	
	if g_currentMission.alternativeTipTrigger ~= nil then
		if not g_currentMission.alternativeTipTrigger.overrideTrailerFruitLimit then
			for fillType in pairs(self.fillTypes) do
				if fillType ~= Fillable.FILLTYPE_UNKNOWN then
					if not g_currentMission.alternativeTipTrigger:isTrailerFruitAllowed(fillType) then
						g_currentMission.alternativeTipTrigger:addTrailerFruit(fillType);
					end;
				end;
			end;
		end;
	end;
	
	self.overrideAltHeapShovelLoad = false;
end;
Shovel.load = Utils.appendedFunction(Shovel.load, AlternativeTippingShovel.shovelLoad);

function AlternativeTippingShovel:delete()
	if g_currentMission.alternativeTipTrigger ~= nil then
		if not g_currentMission.alternativeTipTrigger.overrideTrailerFruitLimit then
			for fillType in pairs(self.fillTypes) do
				if fillType ~= Fillable.FILLTYPE_UNKNOWN then
					if fillType ~= Fillable.FILLTYPE_SUGARBEET and fillType ~= Fillable.FILLTYPE_POTATO and fillType ~= Fillable.FILLTYPE_SILAGE and fillType ~= Fillable.FILLTYPE_MANURE then
						g_currentMission.alternativeTipTrigger:removeTrailerFruit(fillType);
						g_currentMission.alternativeTipTrigger.updateShovel = true;
					end;
				end;
			end;
		end;
	end;
end;
Shovel.delete = Utils.appendedFunction(Shovel.delete, AlternativeTippingShovel.delete);

function AlternativeTippingShovel:shovelUpdate(dt)
	if g_currentMission.alternativeTipTrigger ~= nil and self.altTipPoint ~= nil then
		if self:getIsActive() then
			if self:getIsActiveForInput() then
				if self.movingDirection == 0 and self.fillLevel > 0 then
					if self:getAllowFillShovel(self.currentFillType) then
						if g_currentMission.alternativeTipTrigger:isTrailerFruitAllowed(self.currentFillType) then
							if g_currentMission.alternativeTipTrigger:getTriggerByFilltype(Fillable.fillTypeIntToName[self.currentFillType]) ~= nil then
								local allowed = true;
								local allowed2 = true;
								for _, trigger in pairs(g_currentMission.tipTriggers) do
									if trigger.isExtendedTrigger then
										local x1, y1, z1 = getWorldTranslation(self.altTipPoint);
										local x2, y2, z2 = getWorldTranslation(trigger.triggerId);
										local distance = Utils.vector2Length(x2-x1, z2-z1);
										if trigger.currentFillType ~= self.currentFillType then
											if distance < g_currentMission.alternativeTipTrigger.minDistanceTriggers then
												allowed = false;
											end;
										else
											if distance < g_currentMission.alternativeTipTrigger.minDistanceSameTypeTriggers then
												allowed2 = false;
											end;
										end;
									end;
								end;
								
								self.isInActiveTipRange = not allowed or not allowed2;
								
								if InputBinding.hasEvent(InputBinding.ALTERNATIVETIPPING_START_TIPPING) then
									if allowed then
										if allowed2 then
											g_currentMission.alternativeTipTrigger:addNewTrigger(self, false, AlternativeTipTrigger.SHOVEL, 0, false);
										else
											self.printTriggerError2 = self.time + 3500;
										end;
									else
										self.printTriggerError = self.time + 3500;
									end;
								end;
							end;
						end;
					end;
				end;
				
				if AlternativeTipTrigger.DEBUG then
					local distance = g_currentMission.alternativeTipTrigger.angleMeasureDistance;
					local x, y, z = getWorldTranslation(self.altTipPoint);
					local angleX, th1, th2 = g_currentMission.alternativeTipTrigger:getAngle(x, x, z - distance, z + distance, y, y);
					local angleZ, th3, th4 = g_currentMission.alternativeTipTrigger:getAngle(x + distance, x - distance, z, z, y, y);
					
					if self.isServer then
						drawDebugPoint(x, th1 + 0.1, z - distance, 1, 0, 0, 1);
						drawDebugPoint(x, th2 + 0.1, z + distance, 1, 0, 0, 1);
						drawDebugPoint(x + distance, th3 + 0.1, z, 0, 0, 1, 1);
						drawDebugPoint(x - distance, th4 + 0.1, z, 0, 0, 1, 1);
						
						drawDebugLine(x, y, z, 1, 0, 0, x, th1, z, 1, 0, 0);
						drawDebugLine(x, th1, z, 0, 1, 0, x, th1 - g_currentMission.shovelRayDistance, z, 0, 1, 0);
					end;
					
					g_currentMission:addExtraPrintText("angleX " .. angleX);
					g_currentMission:addExtraPrintText("angleZ " .. angleZ);
					g_currentMission:addExtraPrintText("trailer " .. tostring(self.trailerFound) .. ", supported " .. tostring(self.trailerFoundSupported));
					g_currentMission:addExtraPrintText("object " .. tostring(self.objectFound) .. ", supported " .. tostring(self.objectFoundSupported));
				end;
			end;
		end;
	end;
	
end;
Shovel.update = Utils.appendedFunction(Shovel.update, AlternativeTippingShovel.shovelUpdate);

function AlternativeTippingShovel:shovelUpdateTick(dt)
	if self:getIsActive() and not self.overrideAltHeapShovelLoad then
		local emptySpeed = self:getShovelEmptyingSpeed();
		
		if emptySpeed <= 0 then
			if self.altTipPoint ~= nil then
				self.objectFound = nil;
				-- self.trailerFound = nil;
				
				if self.fillLevel < self.capacity and g_currentMission.shovelRayDistance ~= nil then
					local x, y, z = getWorldTranslation(self.altTipPoint);
					raycastAll(x, y, z, 0, -1, 0, "findTrailerRaycastCallback", Utils.getNoNil(self.altTipRayDistance, g_currentMission.shovelRayDistance), self);
					
					if self.objectFound ~= nil then
						if self.objectFound.isExtendedTrigger then
							self.objectFound:addShovel(self);
						end;
					end;
				end;
			end;
		end;
	end;
	
	if g_currentMission.alternativeTipTrigger ~= nil then
		if not g_currentMission.alternativeTipTrigger.overrideTrailerFruitLimit then
			if g_currentMission.alternativeTipTrigger.updateShovel or g_currentMission.alternativeTipTrigger.forceUpdateShovel then
				for fillType in pairs(self.fillTypes) do
					if fillType ~= Fillable.FILLTYPE_UNKNOWN then
						if not g_currentMission.alternativeTipTrigger:isTrailerFruitAllowed(fillType) then
							g_currentMission.alternativeTipTrigger:addTrailerFruit(fillType);
						end;
					end;
				end;
				
				g_currentMission.alternativeTipTrigger.updateShovel = false;
				g_currentMission.alternativeTipTrigger.forceUpdateShovel = false;
			end;
		end;
		
		if self.fillLevel == 0 and self.currentFillType ~= Fillable.FILLTYPE_UNKNOWN then
			self:setFillLevel(0, Fillable.FILLTYPE_UNKNOWN, true);
		end;
	end;
end;
Shovel.updateTick = Utils.appendedFunction(Shovel.updateTick, AlternativeTippingShovel.shovelUpdateTick);

function AlternativeTippingShovel:draw()
	if self.movingDirection == 0 and g_currentMission.alternativeTipTrigger ~= nil then
		if self:getIsActiveForInput() then
			if self.movingDirection == 0 and self.fillLevel > 0 then
				if self:getAllowFillShovel(self.currentFillType) then
					if g_currentMission.alternativeTipTrigger:isTrailerFruitAllowed(self.currentFillType) then
						if g_currentMission.alternativeTipTrigger:getTriggerByFilltype(Fillable.fillTypeIntToName[self.currentFillType]) ~= nil then
							if not self.isInActiveTipRange then
								g_currentMission:addHelpButtonText(g_i18n:getText("ALTERNATIVETIPPING_START_TIPPING"), InputBinding.ALTERNATIVETIPPING_START_TIPPING);
							end;
						end;
					end;
				end;
			end;
			
			if self.printTriggerError ~= nil then
				if self.time < self.printTriggerError then
					g_currentMission:addWarning(g_i18n:getText("ALTERNATIVETIPPING_TIPPING_NOT_ALLOWED"), 0.018, 0.033);
				end;
			end;
			if self.printTriggerError2 ~= nil then
				if self.time < self.printTriggerError2 then
					g_currentMission:addWarning(g_i18n:getText("ALTERNATIVETIPPING_TIPPING_NOT_ALLOWED2"), 0.018, 0.033);
				end;
			end;
		end;
	end;
end;
Shovel.draw = Utils.appendedFunction(Shovel.draw, AlternativeTippingShovel.draw);

function AlternativeTippingShovel:findTrailerRaycastCallback(superFunc, transformId, x, y, z, distance)
	local trailer = g_currentMission.objectToTrailer[transformId];
	if trailer ~= nil and trailer ~= self then
		if trailer:allowFillType(self.currentFillType) and trailer.getAllowFillFromAir ~= nil and trailer:getAllowFillFromAir() then
			self.trailerFound = trailer;
			self.trailerFoundSupported = true;
		else
			if self.trailerFound == nil then
				self.trailerFound = trailer;
				self.trailerFoundSupported = false;
			end
		end
		
		return false;
	end;
	
	local object = g_currentMission:getNodeObject(transformId);
	if object ~= nil and object ~= self and object.addShovelFillLevel ~= nil and object.getAllowShovelFillType ~= nil then
		local terrainHeight = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x, y, z);
		local offset = Utils.getNoNil(object.shovelTerrainHeapHeightOffset, g_currentMission.shovelTerrainHeapHeightOffset);
		
		if y > (terrainHeight - offset) then	-- only allow unload to heaps above ground level.
			if object:getAllowShovelFillType(self.currentFillType) then
				self.objectFound = object;
				self.objectFoundSupported = true;
			else
				if self.objectFound == nil then
					self.objectFound = object;
					self.objectFoundSupported = false;
				end;
			end;
			
			return false;
		end;
	end;
	return true;
end;
Shovel.findTrailerRaycastCallback = Utils.overwrittenFunction(Shovel.findTrailerRaycastCallback, AlternativeTippingShovel.findTrailerRaycastCallback);