--
-- AlternativeTippingTrailer
--
-- SFM-Modding
-- @author  Manuel Leithner
-- @date:		01/08/11
-- @version:	v1.11
-- @history:	v1.0 - initial implementation
-- 				v1.1 - FS13 Conversion - Xentro (Marcus@Xentro.se)
--				v1.11 - 
--
--

AlternativeTippingTrailer = {};

function AlternativeTippingTrailer:load(xmlFile)
	self.useAngleBased = Utils.getNoNil(getXMLBool(xmlFile, "vehicle.alternativeTipping#useAngleBased"), true);
	self.allowedSpecsAltTip = SpecializationUtil.hasSpecialization(StrawBlower, self.specializations) or SpecializationUtil.hasSpecialization(MixerWagon, self.specializations);
end;
Trailer.load = Utils.appendedFunction(Trailer.load, AlternativeTippingTrailer.load);

function AlternativeTippingTrailer:update(dt)
	if g_currentMission.alternativeTipTrigger ~= nil then
		local tipIndex = Utils.getNoNil(self.toggleTipSideIndex, g_currentMission.alternativeTipTrigger.referenceTipIndex); -- ToggleTipSide compatible
				
		if self.tipReferencePoints[tipIndex] ~= nil then
			if self.triggerPlacement == nil or (self.triggerPlacement ~= nil and self.triggerPlacement ~= self.tipReferencePoints[tipIndex].node) then
				self.triggerPlacement = self.tipReferencePoints[tipIndex].node;
			end;
		end;
		
		if self:getIsActive() and self.triggerPlacement ~= nil then
			if not self.allowedSpecsAltTip then
				self.isInActiveTipRange = g_currentMission.trailerInTipRange == self;
				
				if self:getIsActiveForInput() then
					local allow = false;
					if self.isRealistic then
						allow = math.abs(self.lastSpeedReal) < 0.0002;
					else
						allow = self.movingDirection == 0;
					end;
					
					if allow and not self.isInActiveTipRange and self.fillLevel > 0 and self.tipState == Trailer.TIPSTATE_CLOSED then
						if g_currentMission.alternativeTipTrigger:getTriggerByFilltype(Fillable.fillTypeIntToName[self.currentFillType]) ~= nil then
							if not g_currentMission.alternativeTipTrigger:isTrailerFruitAllowed(self.currentFillType) then
								if Input.isKeyPressed(Input.KEY_lctrl) then
									self.printTriggerError2 = self.time + 500;
								end;
							end;
							
							if InputBinding.hasEvent(InputBinding.ALTERNATIVETIPPING_START_TIPPING) then
								local allowed = true;
								for _, trigger in pairs(g_currentMission.tipTriggers) do
									if trigger.isExtendedTrigger and trigger.currentFillType ~= self.currentFillType then
										local x1, y1, z1 = getWorldTranslation(self.triggerPlacement);
										local x2, y2, z2 = getWorldTranslation(trigger.triggerId);
										local distance = Utils.vector2Length(x2-x1, z2-z1);
										if distance < g_currentMission.alternativeTipTrigger.minDistanceTriggers then
											allowed = false;
											self.printTriggerError = self.time + 3500;
											break;
										end;
									end;
								end;
								
								if allowed then
									g_currentMission.alternativeTipTrigger:addNewTrigger(self, false, AlternativeTipTrigger.TRAILER, tipIndex, false);
								end;
							end;
						end;
					end;
				end;
				
				if AlternativeTipTrigger.DEBUG and self.useAngleBased then
					local distance = g_currentMission.alternativeTipTrigger.angleMeasureDistance;
					local x, y, z = getWorldTranslation(self.triggerPlacement);
					local angleX, th1, th2 = g_currentMission.alternativeTipTrigger:getAngle(x, x, z - distance, z + distance, y, y);
					local angleZ, th3, th4 = g_currentMission.alternativeTipTrigger:getAngle(x + distance, x - distance, z, z, y, y);
					
					if self.isServer then
						drawDebugPoint(x, th1 + 0.1, z - distance, 1, 0, 0, 1);
						drawDebugPoint(x, th2 + 0.1, z + distance, 1, 0, 0, 1);
						drawDebugPoint(x + distance, th3 + 0.1, z, 0, 0, 1, 1);
						drawDebugPoint(x - distance, th4 + 0.1, z, 0, 0, 1, 1);
						
						drawDebugLine(x, y, z, 1, 0, 0, x, th1, z, 1, 0, 0);
					end;
					
					g_currentMission:addExtraPrintText("angleX " .. angleX);
					g_currentMission:addExtraPrintText("angleZ " .. angleZ);
				end;
			else
				self.isInActiveTipRange = true;
			end;
		end;
	end;
end;
Trailer.update = Utils.appendedFunction(Trailer.update, AlternativeTippingTrailer.update);

function AlternativeTippingTrailer:draw()
	if g_currentMission.alternativeTipTrigger:getTriggerByFilltype("extraFruit") ~= nil then
		g_currentMission:addHelpButtonText(g_i18n:getText("ALTERNATIVETIPPING_START_TIPPING") .. " klee" , InputBinding.ALTERNATIVETIPPING_START_TIPPING);
	end;
	
	local allow = false;
	if self.isRealistic then
		allow = math.abs(self.lastSpeedReal) < 0.0002;
	else
		allow =  self.movingDirection == 0;
	end;
	
	if allow and g_currentMission.alternativeTipTrigger ~= nil then
		if self:getIsActiveForInput() then
			if self.isInActiveTipRange ~= nil and not self.isInActiveTipRange and self.fillLevel > 0 and self.tipState == Trailer.TIPSTATE_CLOSED then
				if g_currentMission.alternativeTipTrigger:isTrailerFruitAllowed(self.currentFillType) then
					if g_currentMission.alternativeTipTrigger:getTriggerByFilltype(Fillable.fillTypeIntToName[self.currentFillType]) ~= nil then
						g_currentMission:addHelpButtonText(g_i18n:getText("ALTERNATIVETIPPING_START_TIPPING"), InputBinding.ALTERNATIVETIPPING_START_TIPPING);
					end;
				end;
			end;
			
			if self.printTriggerError ~= nil then
				if self.time < self.printTriggerError then
					g_currentMission:addWarning(g_i18n:getText("ALTERNATIVETIPPING_TIPPING_NOT_ALLOWED"), 0.018, 0.033);
				end;
			end;
			
			if self.fillLevel > 0 then
				if self.printTriggerError2 ~= nil and self.time < self.printTriggerError2 then
					g_currentMission:addWarning(g_i18n:getText("ALTERNATIVETIPPING_TIPPING_WARNING") .. " \n\n" .. string.format(g_i18n:getText("ALTERNATIVETIPPING_TIPPING_WARNING_DESC"), InputBinding.getKeyNamesOfDigitalAction(InputBinding.ALTERNATIVETIPPING_START_TIPPING)), 0.018, 0.033);
				end;
			end;
		end;
	end;
end;
Trailer.draw = Utils.appendedFunction(Trailer.draw, AlternativeTippingTrailer.draw);